local L = Capitalist.Locale
local realm = GetRealmName()
local character = UnitName("player")
local faction = UnitFactionGroup("player")

local defaultProductCount = 1

local productsMemory = { }

function Capitalist:ProductMemoryReset()
	productsMemory = { }
end

function Capitalist:ProductIsEnabled(profession, skillName)
	return (self.db.profile.professions and
			self.db.profile.professions[profession] and
			self.db.profile.professions[profession][skillName] and
			(self.db.profile.professions[profession][skillName].enabled == true))
end

function Capitalist:ProductSetEnabled(profession, skillName, enabled)
	if (enabled) then
		self.db.profile.professions[profession][skillName].enabled = true
	else
		self.db.profile.professions[profession][skillName].enabled = nil
	end
end

function Capitalist:ProductGetCount(profession, skillName)
	local countEffective = 0

	if (self.db.profile.professions and
		self.db.profile.professions[profession] and
		self.db.profile.professions[profession][skillName]) then

		countEffective = self.db.profile.professions[profession][skillName].count
		if (not countEffective) then
			countEffective = defaultProductCount
		end
	end

	return countEffective
end

function Capitalist:ProductSetCount(profession, skillName, count)
	if (not self.db.profile.professions) then
		self.db.profile.professions = {}
	end
	if (not self.db.profile.professions[profession]) then
		self.db.profile.professions[profession] = {}
	end
	if (not self.db.profile.professions[profession][skillName]) then
		self.db.profile.professions[profession][skillName] = {}
	end

	self.db.profile.professions[profession][skillName].count = count
end

function Capitalist:ProductIsAllowed(profession, skillName)
	if ((not self.db.profile.professions) or
		(not self.db.profile.professions[profession]) or
		(not self.db.profile.professions[profession][skillName])) then
		return false
	end
	
	local skillData = self.db.profile.professions[profession][skillName]
	local recipeLink = skillData.recipeLink
		
	if ((not L.expansionRanges[profession]) or (not recipeLink)) then
		return true
	end

	if (L.expansionRangeExceptions[skillName] == "Cataclysm") then
		return self.db.profile.filters.global.Cataclysm
	elseif (L.expansionRangeExceptions[skillName] == "WOTLK") then
		return self.db.profile.filters.global.Northrend
	elseif (L.expansionRangeExceptions[skillName] == "BC") then
		return self.db.profile.filters.global.BurningCrusade
	elseif (L.expansionRangeExceptions[skillName] == "Classic") then
		return self.db.profile.filters.global.Classic
	end

	local recipeIdString = string.match(recipeLink, "Henchant:(%d+)")
	local recipeIdNumber = tonumber(recipeIdString)

	if (recipeIdNumber >= L.expansionRanges[profession]["Cataclysm"]) then
		return self.db.profile.filters.global.Cataclysm
	elseif (recipeIdNumber >= L.expansionRanges[profession]["WOTLK"]) then
		return self.db.profile.filters.global.Northrend
	elseif (recipeIdNumber >= L.expansionRanges[profession]["BC"]) then
		return self.db.profile.filters.global.BurningCrusade
	else
		return self.db.profile.filters.global.Classic
	end
end

function Capitalist:ProductIsVisible(profession, skillName)
	return (not productsMemory[profession]) or
		   (not productsMemory[profession][skillName]) or
		   (not productsMemory[profession][skillName].hidden)
end

function Capitalist:ProductSetVisible(profession, skillName, visible)
	if (visible) then
		if (productsMemory[profession]) then
			if (productsMemory[profession][skillName]) then
				productsMemory[profession][skillName].hidden = nil
			end
		end
	else
		if (not productsMemory[profession]) then
			productsMemory[profession] = {}
		end
		if (not productsMemory[profession][skillName]) then
			productsMemory[profession][skillName] = {}
		end
		productsMemory[profession][skillName].hidden = true
	end
end

function Capitalist:FindProfession(skillName)
	if (self.db.profile.professions) then
		for profession, professionData in pairs(self.db.profile.professions) do
			if (professionData[skillName]) then
				return profession
			end
		end
	end
	return nil
end

function Capitalist:FindCrafter(profession, skillName)
	local oRealm = self:GetProfile_Realm_Faction(realm, faction)
	if (oRealm) then
		local crafterFound
		for crafter, crafterData in pairs(oRealm) do
			if (crafterData[profession] and crafterData[profession][skillName]) then

				-- prefer current character, when we have the option
				if ((not crafterFound) or (crafter == character)) then
					crafterFound = crafter
				end
			end
		end
		return crafterFound
	end
	return nil
end

-- REVIEW: Blizzard messed up the product names for some of the scrolls.
--         Hardcoded corrections here.
function Capitalist:ApplyScrollProductNameCorrection(productName)
	local productNameCorrected = productName
	if (productNameCorrected == "Enchant Shield - Greater Intellect") then
		productNameCorrected = "Shield Enchant - Intellect III"
	end
	return productNameCorrected
end

function Capitalist:ApplyGlyphProductNameCorrectionViaId(productName, itemId)
	local productNameCorrected = productName

	-- item 45804, spell 64266 - DK
	if (itemId == 45804) then
		productName = string.gsub(productName, "Death Coil", "Death Coil <Death Knight>")
	end
	-- item 42457, spell 57261 - Warlock
	if (itemId == 42457) then
		productName = string.gsub(productName, "Death Coil", "Death Coil <Warlock>")
	end

	return productName
end

function Capitalist:ApplyGlyphProductNameCorrection(productName, itemLink)
	local itemIdString = string.match(itemLink, "Hitem:(%d+)")
	local itemIdNumber = tonumber(itemIdString)
	return self:ApplyGlyphProductNameCorrectionViaId(productName, itemIdNumber)
end

function Capitalist:ProductTextFromSkillName(profession, skillName)
	local itemLink = self.db.profile.professions[profession][skillName].itemLink
	if (itemLink) then

		-- REVIEW: It's safe to use the ProductName function here because it uses gsub instead of appending directly
		itemLink = self:ApplyGlyphProductNameCorrection(itemLink, itemLink)

		return itemLink
	else
		return self:ProductNameFromSkillName(profession, skillName)
	end
end

function Capitalist:SkillTextFromSkillName(profession, skillName)
	local recipeLink = self.db.profile.professions[profession][skillName].recipeLink
	if (recipeLink) then

		-- REVIEW: It's safe to use the ProductName function here because it uses gsub instead of appending directly
		recipeLink = self:ApplyGlyphProductNameCorrection(recipeLink, recipeLink)

		return recipeLink
	else
		return skillName
	end
end

local function ColorizeForSkillType(input, skillType)
	if (skillType == "trivial") then
		return "|cff808080"..input.."|r"
	elseif (skillType == "easy") then
		return "|cff00ff00"..input.."|r"
	elseif (skillType == "medium") then
		return "|cffffff00"..input.."|r"
	elseif (skillType == "optimal") then
		return "|cffff8000"..input.."|r"
	elseif (skillType == "difficult") then
		return "|cffff8080"..input.."|r"
	else
		return input
	end
end

function Capitalist:OutputProductLink(profession, prefix, skillName, count, crafter)
	if (self.db.profile.professions and
		self.db.profile.professions[profession] and
		self.db.profile.professions[profession][skillName]) then

		local itemText = self:ProductTextFromSkillName(profession, skillName)

		local s = self:StringForItem(prefix, itemText, count)
		if (self.db.profile.professions[profession][skillName].min) then
			s = s.." ("..self.db.profile.professions[profession][skillName].min
			if (self.db.profile.professions[profession][skillName].max) then
				s = s.."-"..self.db.profile.professions[profession][skillName].max
			end
			s = s..")"
		end

		local cost = self:CalculateSkillCopperCost(profession, skillName, false)
		if (cost) then
			s = s.." @ "..self:PrettyCurrency(cost).." ea."
		end		

		if (crafter) then
			local oSkill = self:GetProfile_Realm_Faction_Character_Profession_Skill(realm, faction, crafter, profession, skillName)
			local skillType = oSkill.type
			s = s.." by "..ColorizeForSkillType(crafter, skillType)
		end
		s = s.." with "..self:SkillTextFromSkillName(profession, skillName)
		
		self:Print(s)
	end
end

function Capitalist:FindProducts(...)
	local pattern = ...
	if (self.db.profile.professions) then
		for profession, professionData in pairs(self.db.profile.professions) do
			for skillName, skillData in pairs(professionData) do
				if (strlower(skillName):find(strlower(pattern))) then
					self:OutputProductLink(profession, "", skillName, 1)
				end
			end
		end
	end
end

function Capitalist:OutputProductCraftingSequence(...)
	local productName, count, prefix = ...
	if (not count) then
		count = 1
	end
	if (not prefix) then
		prefix = ""
	end

	local skills = self:SkillsFromProductName(productName)
	if (skills) then
		for skillName, profession in pairs(skills) do
			local crafter = self:FindCrafter(profession, skillName)
			self:OutputProductLink(profession, prefix, skillName, count, crafter)

			if (crafter) then
				for componentName, componentCount in pairs(self.db.profile.professions[profession][skillName].components) do
					self:OutputProductCraftingSequence(componentName, componentCount, prefix.."    ")
				end
			end
		end
	elseif (self:IsResource(productName)) then
		self:OutputComponent(prefix, productName, count)
	else
		self:Print(string.format(L["Product %s unknown."], productName))
	end
end

function Capitalist:ReportTop10(...)
	self:DebugEntry("ReportTop10")

	local tableSorted = {}

	local oCharacter = self:GetProfile_Realm_Faction_Character(realm, faction, character)
	if (oCharacter) then
		for profession, professionData in pairs(oCharacter) do

			-- Ignore salvage data
			if (self.db.profile.professions[profession]) then

				for skillName, skillData in pairs(professionData) do
					local numCraft = skillData.numCraft or 0
					if (numCraft > 0) then
						local entry = { c = numCraft, n = skillName }
						table.insert(tableSorted, entry)
					end
				end
			end
		end
	end

	-- REVIEW: Add logic for salvage skills like UpdateStatistics in options.lua does

	table.sort(tableSorted, function(a, b) return a.c > b.c end)

	for i=1, math.min(#tableSorted, 10) do
		local e = tableSorted[i]
		self:Print(e.c.."x "..e.n)
	end

	self:DebugExit("ReportTop10")
end


function Capitalist:ResetSkillCaches()
	self:ResetProductsFromSkills()
	self:UpdateCooldownSkills()
end



Capitalist.cooldownSkills = { }

function Capitalist:UpdateCooldownSkills()
	self.cooldownSkills = { }
	if (self.db.profile.professions) then
		for profession, professionData in pairs(self.db.profile.professions) do
			self.cooldownSkills[profession] = { }
			for skillName, skillData in pairs(professionData) do
				-- TODO: Arcanite Bar
				-- REVIEW: Exclude transmutes to avoid infinite recursion (earth -> fire -> earth ...)
				if (self.db.profile.professions[profession][skillName].cooldownApplies) then
-- TODO: remove next line
--				if (skillName:find(L["Transmute:"])) then
					table.insert(self.cooldownSkills[profession], skillName)
				end
			end
		end
	end
end

function Capitalist:SkillCooldownApplies(profession, skillName)
	if (self.cooldownSkills and
		self.cooldownSkills[profession] and
		tContains(self.cooldownSkills[profession], skillName)) then

		return true
	end
	return false
end

-- TODO: mappingtables.lua ?

Capitalist.mapProductsFromSkills = { }
Capitalist.mapSkillsFromProducts = { }

function Capitalist:ResetProductsFromSkills()
	self.mapProductsFromSkills = { }
	if (self.db.profile.professions) then
		for profession, professionData in pairs(self.db.profile.professions) do
			self.mapProductsFromSkills[profession] = { }
		end
	end
end
function Capitalist:ResetSkillsFromProducts()
	self.mapSkillsFromProducts = { }
end

function Capitalist:ProductNameFromSkillName(profession, skillName)
	local productName
	if (self.mapProductsFromSkills[profession]) then
		productName = self.mapProductsFromSkills[profession][skillName]
		if (not productName) then
			local itemLink = self.db.profile.professions[profession][skillName].itemLink
			if (itemLink) then
				productName = itemLink:match("%[(.*)%]")

				if (productName:find(L["Enchant "])) then
					productName = self:ApplyScrollProductNameCorrection(productName)
				end
				if (productName:find(L["Glyph of "])) then
					productName = self:ApplyGlyphProductNameCorrection(productName, itemLink)
				end
				
				self.mapProductsFromSkills[profession][skillName] = productName
			end
		end
	end

	return productName
end


function Capitalist:SkillsFromProductName(productName)
	local skills = self.mapSkillsFromProducts[productName]
	if (not skills) then
		if (self.db.profile.professions) then
			for profession, professionData in pairs(self.db.profile.professions) do
				for skillName, skillData in pairs(professionData) do
					if (not self:SkillCooldownApplies(profession, skillName)) then
						if (productName == self:ProductNameFromSkillName(profession, skillName)) then
							if (not skills) then
								skills = {}
							end
							skills[skillName] = profession
						end
					end
				end
			end
		end

		self.mapSkillsFromProducts[productName] = skills
	end

	return skills
end

function Capitalist:SkillNameFromProductName(productName)
	-- TODO: add criteria (lowest crafting price or whatever)
	local skills = self:SkillsFromProductName(productName)
	if (skills) then
		for skillName, profession in pairs(skills) do
			return skillName, profession
		end
	end
	return nil, nil
end

function Capitalist:IsProduct(productName)
	local skills = self:SkillsFromProductName(productName)
	return (skills ~= nil)
end

function Capitalist:ComponentsForSkill(profession, skillName)
	local components

	local product = self.db.profile.professions[profession][skillName]
	if (product) then
		components = {}
		for componentName, componentCount in pairs(product.components) do
			components[componentName] = componentCount
		end
	end

	return components
end

function Capitalist:ComponentsForSkillProduct(profession, skillName)
	local components = self:ComponentsForSkill(profession, skillName)
	if (components) then
		-- enchanting-specific
		local productName = self:ProductNameFromSkillName(profession, skillName)
		if (productName:find(L["Enchant "])) then
			components[L["Enchanting Vellum"]] = 1
		end
	end
	return components
end
